"""
    This file exposes the main API for this library, the microscope Manager
"""
# Local imports
from c4v.scraper.persistency_manager.base_persistency_manager import (
    BasePersistencyManager,
)
from c4v.scraper.persistency_manager.sqlite_storage_manager import SqliteManager
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData, Sources
from c4v.scraper.scraper import bulk_scrape, _get_scraper_from_url
from c4v.scraper.settings import INSTALLED_CRAWLERS
from c4v.classifier.classifier_experiment import ClassifierExperiment
from c4v.classifier.classifier import Classifier
from c4v.classifier.language_model.language_model import LanguageModel
from c4v.config import settings
from c4v.microscope.metadata import Metadata

# Python imports
from typing import Dict, List, Iterable, Callable, Tuple, Any, Union
from pathlib import Path
import sys


class Manager:
    """
        This object encapsulates shared behavior between our multiple components,
        allowing easy access to common operations.
        Parameters:
            persistency_manager : BasePersistencyManager = persistency manager object to use when accessing persistent data storage
            metadata : Metadata = Persistent configuration data
    """

    def __init__(
        self,
        persistency_manager: BasePersistencyManager,
        metadata: Metadata,
        local_files_path: str = settings.c4v_folder,
    ):
        self._persistency_manager = persistency_manager
        self._metadata = metadata
        self._local_files_path = local_files_path

    @property
    def local_files_path(self) -> str:
        """
            Path where files (.c4v folder) will be saved
        """
        return self._local_files_path

    @local_files_path.setter
    def local_files_path(self, new_path: str):
        self._local_files_path = new_path

    @property
    def persistency_manager(self) -> BasePersistencyManager:
        return self._persistency_manager

    def get_bulk_data_for(
        self, urls: List[str], should_scrape: bool = True, save: bool = True
    ) -> List[ScrapedData]:
        """
            Retrieve scraped data for given url set if scrapable
            Parameters:
                urls : [str] = urls whose data is to be retrieved. If not available yet, then scrape it if requested so
                should_scrape : bool = (optional) if should scrape non-existent urls
                save : bool = (optional) if should save new data when scraped for the first time
        """
        # just a shortcut
        db = self._persistency_manager

        # Separate scraped urls from non scraped
        not_scraped = db.filter_scraped_urls(urls)

        # Scrape missing instances if necessary
        if should_scrape and not_scraped:
            items = bulk_scrape(not_scraped)
            if save:
                db.save(items)  # save if requested to

        # Convert to set to speed up lookup
        urls = set(urls)
        return [sd for sd in db.get_all() if sd.url in urls]

    def get_data_for(self, url: str, should_scrape: bool = True) -> ScrapedData:
        """
            Get data for this url if stored and scrapable. May return none if could not
            find data for this url
            Parameters:
                url : str = url to be scraped
                should_scrape : bool = if should scrape this url if not available in db
        """
        data = self.get_bulk_data_for([url], should_scrape)
        return data[0] if data else None

    def scrape_pending(self, limit: int = -1):
        """
            Update DB by scraping rows with no scraped data, just the url
            Parameters:
                limit : int = how much measurements to scrape, set a negative number for no limit
        """

        db = self._persistency_manager

        scrape_urls = [d.url for d in db.get_all(limit=limit, scraped=False)]

        scraped = bulk_scrape(scrape_urls)

        db.save(scraped)

    def get_all(self, limit: int = -1, scraped: bool = None) -> Iterable[ScrapedData]:
        """
            Return all ScrapedData instances available, up to "limit" rows. If scraped = true, then
            return only scraped instances. If scraped = false, return only non scraped. Otherwise, return anything
            Parameters:
                limit : int = max ammount of rows to return. If negative, there's no limit of rows
                scraped : bool = If instances should be scraped or not. If true, all of them should be scraped. If false, none of 
                                them should be scraped. If None, there's no restriction
            Return:
                An iterator returning available rows
        """
        return self._persistency_manager.get_all(limit, scraped)

    def crawl_new_urls_for(
        self,
        crawler_names: List[str] = None,
        post_process: Callable[[List[str]], None] = None,
        limit=-1,
    ):
        """
            Crawl for new urls using the given crawlers only
            Parameters:
                crawler_names : [str] = names of crawlers to be ran when this function is called. If no list is passed, then 
                                        all crawlers will be used
                post_process : ([str]) -> None = Function to call over new elements as they come
                limit        : int = Max amount of urls to save, -1 when no limit 
        """
        db = self._persistency_manager
        limit = limit if limit >= 0 else sys.maxsize

        class Counter:
            def __init__(self):
                self.count = 0

            def add(self, x: int):
                self.count += x

        counter = Counter()

        # Function to process urls as they come
        def save_urls(urls: List[str]):
            # Filter already known urls and take at the must the necessary ones to fill the required size
            urls = db.filter_known_urls(urls)[: limit - counter.count]
            datas = [ScrapedData(url=url, source=Sources.SCRAPING) for url in urls]
            db.save(datas)

            # Update how much elements have beed added so far
            counter.add(len(datas))

            # Call any callback function
            if post_process:
                post_process(urls)

        # Function to tell if the crawling process should stop
        def should_stop() -> bool:
            return counter.count >= limit

        # Names for installed crawlers
        crawlers = [c.name for c in INSTALLED_CRAWLERS]

        # if no list provided, default to every crawler
        if crawler_names == None:
            crawler_names = crawlers

        # Instantiate crawlers to use
        crawlers_to_run = [
            crawler() for crawler in INSTALLED_CRAWLERS if crawler.name in crawler_names
        ]

        # crawl for every crawler
        for crawler in crawlers_to_run:
            crawler.crawl_and_process_urls(save_urls, should_stop)

    def split_non_scrapable(self, urls: List[str]) -> Tuple[List[str], List[str]]:
        """
            splits url list in two list, one having a list of scrapable urls, 
            and another one with only non-scrapable
            Parameters:
                urls : [str] = List of urls to be split
            Return:
                A tuple with two lists, the first one with only scrapable urls from input list,
                the second one with non-scrapable ones
        """
        scrapable, non_scrapable = [], []

        for url in urls:
            try:
                _get_scraper_from_url(url)
                scrapable.append(url)
            except ValueError:
                non_scrapable.append(url)

        return scrapable, non_scrapable

    @classmethod
    def from_local_sqlite_db(cls, db_path: str, metadata: Union[str, Metadata]):
        """
            Create a new instance using an SQLite local db
            metadata : str | Metadata = Path to file with metadata or metadata instance itself
        """
        if isinstance(metadata, Metadata):
            pass  # Everything ok, just keep going
        elif isinstance(metadata, str):
            # If string is provided, interpret it as a filepath
            metadata = Metadata.from_json(metadata)
        else:
            raise TypeError(
                f"metadata field should be path to metadata files or metadata instance. Provided object's type: {type(metadata)}"
            )

        db = SqliteManager(db_path)
        return cls(db, metadata)

    @classmethod
    def from_default(
        cls, db_path: str = None, metadata: str = None, local_files_path: str = None
    ):
        """
            Create a Manager instance using files from the default `C4V_FOLDER`
        """
        # Set up db
        db = SqliteManager(
            db_path
            or str(
                Path(
                    local_files_path or settings.c4v_folder,
                    settings.local_sqlite_db_name,
                )
            )
        )

        # Set up metadata
        metadata = Metadata.from_json(metadata) if metadata else Metadata()

        return cls(db, metadata, local_files_path or settings.c4v_folder)

    def run_classification_from_experiment(
        self, branch: str, experiment: str, data: List[ScrapedData]
    ) -> List[Dict[str, Any]]:
        """
            Classify given data instance list, returning its metrics
            Parameters:
                branch : str = branch name of model to use
                experiment : str = experiment name storing model
                data : [ScrapedData] = Instance to be classified
            Return:
                A List of dicts with the resulting scraped data correctly labelled
                and its corresponding scores tensor for each possible label. Available fields:
                    + data : ScrapedData = resulting data instance after classification
                    + scores : torch.Tensor = Scores for each label returned by the classifier
        """

        classifier_experiment = ClassifierExperiment.from_branch_and_experiment(
            branch, experiment
        )

        return classifier_experiment.classify(data)

    def run_pending_classification_from_experiment(
        self, branch: str, experiment: str, save: bool = True, limit: int = -1
    ) -> List[Dict[str, Any]]:
        """
            Classify data pending for classification in local db, returning the obtained results and saving it 
            to database. 
            Parameters:
                branch : str = branch name
                experiment : str = experiment name
                save : bool = if should store results in db
                limit : maximum number of rows to classify
            Return:
                A List of dicts with the resulting scraped data correctly labelled
                and its corresponding scores tensor for each possible label. Available fields:
                    + data : ScrapedData = resulting data instance after classification
                    + scores : torch.Tensor = Scores for each label returned by the classifier
        """
        # Parse limit
        limit = limit if limit >= 0 else sys.maxsize

        # Request at the most "limit" instances
        data = list(
            x for x in self.persistency_manager.get_all(scraped=True) if not x.label
        )[:limit]

        # classify
        results = self.run_classification_from_experiment(branch, experiment, data)

        # Save if requested so
        if save:
            self.persistency_manager.save((x["data"] for x in results))

        return results

    def explain_for_experiment(
        self,
        branch: str,
        experiment: str,
        sentence: str,
        html_file: str = None,
        additional_label: str = None,
    ) -> Dict[str, Any]:
        """
            Explain a sentence using the given branch and experiment
            Parameters:
                branch : str = branch name of model to use
                experiment : str = experiment name storing model
                sentence : str = sentence to explain
                additional_label : str = Label to include in expalantion. If the predicted 
                                         label is different from this one, then explain how 
                                         much this label was contributing to its corresponding value, 
                                         ignored if not provided
            Return:
                Dict with explaination data
        """

        classifier_experiment = ClassifierExperiment.from_branch_and_experiment(
            branch, experiment
        )

        return classifier_experiment.explain(
            sentence, html_file, additional_label=additional_label
        )

    def get_classifier_labels(self) -> List[str]:
        """
            Get list of possible labels for classifier
            Return:
                List with possible output labels for the classifier
        """
        return Classifier.get_labels()

    def should_retrain_base_lang_model(
        self,
        lang_model: LanguageModel,
        db_manager: BasePersistencyManager = None,
        eval_dataset_size: int = 250,
        min_loss: float = settings.default_lang_model_min_loss,
        should_retrain_fn: Callable[[float], bool] = None,
        fields: List[str] = ["title"],
    ) -> bool:
        """
            If should retrain a base language model based on its loss on the given dataset. If a persistency manager is provided, 
            use it instead of the configured one
            Parameters:
                lang_model : LanguageModel = Language Model to reevaluate
                db_manager : BasePersistency manager = (optional) Persistency manager object to retrieve data to use when evaluating
                eval_dataset_size : int = (optional) Size of the dataset to use when evaluating the model
                min_loss : float = (optional) Minimum acceptable loss, if the computed loss is greater than this treshold, returns true. Otherwise returns false
                should_retrain_fn : (float) -> bool = (optional) Function to check if, based on the loss of the model, it should be retrained. 
                                                    Should receive the loss and return a boolean telling if it should be retrained. 
                                                    Providing this argument will override min_loss argument
                fields : [str] = (optional) fields from scraped data instances to use during evaluation
            Return:
                If the given model should be retrained
        """
        # Sanity check
        assert eval_dataset_size > 0, "Eval dataset size should be a possitive number"
        assert min_loss >= 0, "min loss should be non negative"

        # set up retrain function
        should_retrain_fn = should_retrain_fn or (lambda x: x > min_loss)

        # Set up db_manager
        db_manager = db_manager or self._persistency_manager

        # TODO should order by newer, but can't do it right now as we can't ensure
        # news dates can be parsed into datetime objects
        scraped_data = list(db_manager.get_all(limit=eval_dataset_size, scraped=True))
        if not scraped_data:
            raise ValueError(
                "No ScrapedData available for evaluation. You can get more data by using the `c4v scrawl` and `c4v scrape` functions, or using the microscope.Manager object"
            )

        # Set up dataset
        ds = LanguageModel.to_pt_dataset(
            lang_model.create_dataset_from_scraped_data(scraped_data, fields)
        )
        del scraped_data

        # Compute loss
        loss = lang_model.eval_accuracy(ds)
        return should_retrain_fn(loss)



from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import TfidfVectorizer
import pandas as pd
import numpy as np


class BratDataLoader:
    def __init__(self, files_path: list, binary=True):
        """
        Receives filepath of Annotation File and text file.
        These files must be in the same folder.
        Run sef.preprocess() to have train test split vectorized data. 
        ----------------------------------------------------
        Params
            files_path: list. Path to text and annotation files without their file extension.
                It is assumed that both files have the same nanme.
            
            binary: bool. Create one-hot encoded labels. 
                If false, parsed_df contains the counts of each entity per document.

        Examples

	    # Append root folder
	    import sys
	    sys.path.append('../')

	    # Import libraries
	    import numpy as np
	    import pandas as pd
	    import sys
	    from src.c4v.data.data_loader import BratDataLoader

	    # Instantiate class with one dataset
	    loaded_data = BratDataLoader(['../data/processed/brat/sampled_58_30'], binary=True)

	    ## Get parsed dataframe with text and one-hot encoded responses
	    loaded_data.get_parsed_df()

	    # Create training and test sets
	    loaded_data.preprocess()

	    ## Print training and test sets
	    print(
	    f'''
	    Training set shape:
	    X_train: {loaded_data.X_train.shape}
	    y_train: {loaded_data.y_train.shape}

	    Testing set shape:
	    X_test: {loaded_data.X_test.shape}
	    y_test: {loaded_data.y_test.shape}
    ''')            

        """

        # pair_names is a list containing the paths of the annotated pair files, since .txt and .ann have the same
        #
        self.pair_names = files_path

        self.parsed_df = self.get_parsed_df()
        self.__tweets = self.parsed_df.iloc[:, 0]

        self.X = self.parsed_df.iloc[:, 0]
        # TODO: Investigate why labels "gasoline" has float instead of integer types; and "gas" contains NAN
        #  instead of 0 and floats instead of integer.  Start by looking at self.parsed_df
        self.y = (
            self.parsed_df.iloc[:, 1:].fillna(value=0).astype(int)
        )  # replaces all NAN by 0

        # Training (X,y) and Testing (X,y)
        self.X_train = None
        self.y_train = None
        self.X_test = None
        self.y_test = None
        self.split_random_state = None

    def source_as_list(self) -> list:
        return self.pair_names

    def annotation_parser(self, dir_ann_file, print_grouped_annotations=False):
        """
            Helper function to parse Brat's annotation file (.ann). 
                Returns a dataframe with the following schema: 
                    | id                	| object 	|
                    | annotation        	| object 	|
                    | text              	| object 	|
                    | id_parsed         	| object 	|
                    | annotation_parsed 	| object 	|
        --------------------------------------------------------------
        
        Params:
            
            dir_ann_file: String, Default = None. 
                Path to ann file.
            
            print_grouped_annotation: Bool. Default = False. 
                Prints count of Entities and Attributes.
        """

        ### Read Data
        to_parse_df = pd.read_csv(dir_ann_file, sep="\t", header=None)

        # Rename columns
        to_parse_df.columns = ["id", "annotation", "text"]

        # Remove the ID numbers to know if it's an entity (T) or Attribute (A)
        to_parse_df["id_parsed"] = to_parse_df.id.str.replace("\d", "")

        # Remove text span and IDs (T & A) from column. This columns has the name of the attributes and entities
        to_parse_df["annotation_parsed"] = to_parse_df.annotation.str.replace(
            "[\dTA]", ""
        )

        # Remove Relation tags
        # Change Relation Id to Null
        to_parse_df.id_parsed.replace("R", np.nan, inplace=True)

        # Remove nulls
        to_parse_df.dropna(subset=["id_parsed"], inplace=True)

        # Group by id_parsed, annotation parsed and count results
        df = (
            to_parse_df[["id_parsed", "annotation_parsed"]]
            .groupby(["id_parsed", "annotation_parsed"], sort=True)
            .agg({"annotation_parsed": ["count"]})
            .copy()
        )

        # After the group by there's multi-index columns. We rename the columns to have the level that we want (count)
        df.columns = df.columns.levels[1]

        # sort_values by index. Here the trick is to also use sort_index!!

        if print_grouped_annotations == True:

            print(
                df.sort_values("count", ascending=False).sort_index(
                    level=[0], ascending=[True]
                )
            )
        else:
            pass

        return to_parse_df

    def annotation_merger(self, path_to_txt_file, path_to_ann_file):
        """
        Helper function to merge text file and annotation file created from Brat. 

        The purpose of this function is to flatten the annotations in respect to the text. 
        This function only takes into account entities. Attributes and relations are not considered. 
        The output of this function is intended to serve as input for baseline Machine Learning algorithm.

        The function returns a Pandas Data Frame with the following schema:

        | # | Column     | Non-Null Count | Dtype  |   |
        |---|------------|----------------|--------|---|
        | 0 | text       | n non-null     | object |   |
        | 1 | <entity_1> | n non-null     | uint8  |   |
        | 2 | <entity_n> | n non-null     | uint8  |   |

        The first column corresponds to the text where the tags were made.
        The following <entity> columns are the count of the each entity tagged in the text.

        --------------------------------------------------------------------------------------------------

        Params

        path_to_txt_file: String, default = none.
        Complete or relative path to text file where annotations were made.

        path_to_ann_file: String, default = none.
        Complete or relative path to annotation file (.ann) where the annotations were stored. 

        """

        # Read sampled data

        sampled_ann = self.annotation_parser(dir_ann_file=path_to_ann_file)

        # Subset Entities and rewrite dataframe
        sampled_ann = sampled_ann[sampled_ann.id_parsed == "T"]

        # Create span columns. Split by space.
        split_ann = sampled_ann.annotation.str.split(" ", expand=True)

        # Rename Columns
        split_ann.columns = ["Entities", "first_char", "last_char"]

        # Create new columns with each annotation's text.
        split_ann["text"] = sampled_ann.loc[sampled_ann.id_parsed == "T", "text"]

        split_ann.first_char = split_ann.first_char.astype(int)
        split_ann.last_char = split_ann.last_char.astype(int)

        with open(path_to_txt_file) as f:

            lines = f.readlines()

            # save number line, length of the text and text without break lines: /n
            # assuming one line corresponds to a single tweet
            tuple_tweets = [(len(line), line) for line in lines if len(line) > 0]

            start, end, text_ = list(), list(), list()
            new_start = 0
            for ttw in tuple_tweets:

                # adds the length of the tweet
                start.append(new_start)

                # finds the location of the last character of the tweet
                end.append(new_start + ttw[0] - 1)
                text_.append(ttw[1])

                # gets the starting position of the next tweet
                new_start = new_start + ttw[0]

            text_df = pd.DataFrame(
                {"first_char": start, "last_char": end, "text": text_}
            )

        # Initialize column multi_id
        split_ann["multi_id"] = 0

        # For each row of original text
        for i in range(text_df.shape[0]):

            # Match first character position of the annotation greater or equal to the first_char of the tweet
            # Match last character position of the annotation less than the last character of the complete tweet
            # This will subset the annotations on the complete span of the tweet.
            idx = split_ann[
                (split_ann.first_char >= text_df.loc[i, "first_char"])
                & (split_ann.first_char < text_df.loc[i, "last_char"])
            ].index

            #     Create common value for each dataframe.
            split_ann.loc[idx, "multi_id"] = i
            text_df.loc[i, "multi_id"] = i

        # Merge both dataframes and select the entities and the tweets' complete text
        merged = pd.merge(split_ann, text_df, on="multi_id").loc[
            :, ["Entities", "text_y"]
        ]

        merged.columns = ["entities", "text"]
        # Create dummy variables from entities (these are the tags)
        dummies = pd.get_dummies(merged.entities)

        # Group by the complete text and reset index to flatten multi-index columns
        return (
            pd.concat([merged["text"], dummies], axis=1)
            .groupby("text")
            .sum()
            .reset_index()
        )

    def get_parsed_df(self, binary=True) -> pd.DataFrame:
        """
        Parses DataFrame with helper function annotation_merger.
        Merges annotated entities and text file created from BRAT annotation package. 
        It is assumed that each text file's line correspond to a different document (e.g. Tweet). 

        Returns DataFrame with one-hot encoded labels if binary = True. 
        If binary = False, returns DataFrame with label counts. 


        """

        # self.parsed_df = self.annotation_merger(path_to_txt_file=f'{self.pair_names[0]}.txt',
        #                                         path_to_ann_file=f'{self.pair_names[0]}.ann')
        #

        single_df_list = [
            self.annotation_merger(
                path_to_txt_file=f"{path}.txt", path_to_ann_file=f"{path}.ann"
            )
            for path in self.pair_names
        ]

        merged_parsed_df = pd.concat(single_df_list, axis=0).reset_index(drop=True)

        if binary == True:

            merged_parsed_df.iloc[:, 1:] = merged_parsed_df.iloc[:, 1:].mask(
                merged_parsed_df.iloc[:, 1:] > 1, 1
            )

            return merged_parsed_df

        else:

            return merged_parsed_df

    def preprocess(
        self, test_size=0.3, random_state=21, to_vectorize=True, method="tfidf"
    ):
        """
        Creates Train Test Split.

        If vectorize = True, X_train, X_test, and X will be vectorized.
        If vectorize = False, X_train, X_test, and X will remain as raw_text.

        """

        self.split_random_state = random_state
        if to_vectorize is True:
            self.vectorize()
        else:
            pass

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self.X, self.y, test_size=0.3, random_state=random_state
        )

    def get_X_as_text(self) -> tuple:
        """
        Returns the input data as text, instead of vectors

        :return: A tuple with full records of input as text on the first position; records for training as text
          on the second position; and records used for testing in the third position.
        """

        train, test = train_test_split(
            self.__tweets, test_size=0.3, random_state=self.split_random_state
        )
        return self.__tweets, train, test

    def vectorize(
        self,
        method="tfidf",
        stop_words=None,
        strip_accents="unicode",
        analyzer="word",
        ngram_range=(1, 3),
        norm="l2",
        max_features=10000,
    ):
        """
        Computes specified vectorization method and stores it into self.X. 
        Do not run after pre-process method
        """

        if method == "tfidf":

            tfidf_vectorizer = TfidfVectorizer(
                strip_accents=strip_accents,
                analyzer=analyzer,
                ngram_range=ngram_range,
                norm=norm,
                max_features=max_features,
            )

            self.X = tfidf_vectorizer.fit_transform(self.X)

        else:
            print("Select a vectorization method from the available options")



"""
    An experiment represents a process ran by demand by a developer through scripting,
    which may need some special files and folders, and access to other
    components of the c4v library, the PersistencyManager class for example. 
    Also, experiments are provided with a folder for their necessary files.

    Experiments should provide a summary for results, and an Args Object to use 
    when they run.
"""
# Python imports
import pytz
import sys
import shutil
from pathlib import Path
from datetime import datetime
from typing import List, Type
from dataclasses import dataclass, field


# Local imports
from c4v.config import settings


@dataclass
class BaseExperimentSummary:
    """
        Provide an output summary describing results for this experiment, 
        and an human readable representation.
        
        Note that in order to inherit this class, you must provide default 
        values for every field (required by dataclasses). If you want
        to mark some fields as mandatory, you can do so in the __post_init__
        method.
    """

    # Initialize to date when created
    date: datetime = field(default_factory=lambda: datetime.now(tz=pytz.utc))

    # Optional description
    description: str = None

    def __str__(self) -> str:
        # An human readable representation
        output = f"Date: {datetime.strftime( self.date, settings.date_format )}\n"
        output += f"Description: {self.description or '<no description>'}\n"

        return output


@dataclass
class BaseExperimentArguments:
    """
        Arguments to pass to an experiment run
    """

    pass


class ExperimentFSManager:
    """
        This class manages files for experiments and storing results properly, 
        so experiments doesn't have to worry about where their files will be saved and 
        how their files are managed.

        Experiments are (usually) stored in the 'experiments' folder, under the c4v folder,
        typically in: $HOME/.c4v/experiments.

        Every Experiment has a branch name and an experiment name, the folder provided to every 
        experiment is <experiments_folder>/<branch_name>/<experiment_name>
    """

    _experiments_folder = settings.experiments_dir

    def __init__(
        self, branch_name: str, experiment_name: str, experiments_folder: str = None
    ):

        self._branch_name = branch_name
        self._experiment_name = experiment_name

        # Set up experiments
        self._set_up_experiments_folder(experiments_folder)

    def _set_up_experiments_folder(self, custom_folder: str = None):
        """
            Set up experiment folder, if custom folder is provided, 
            use that folder and assume it does exists, otherwise, 
            create a default one
        """

        # if custom folder provided, just use it assuming it does exist
        if custom_folder:
            self._experiments_folder = custom_folder
            return

        # Set up experiments folder
        experiments_folder_path = Path(settings.c4v_folder, "experiments/")

        # Create folder if not exists
        if not experiments_folder_path.exists():
            experiments_folder_path.mkdir(parents=True)

        self._experiments_folder = str(experiments_folder_path)

        # Set up experiments path
        experiment_content_path = Path(self.experiment_content_folder)
        if not experiment_content_path.exists():
            experiment_content_path.mkdir(parents=True)

    def _get_or_create_path_to(self, folder: str) -> str:
        """
            Get path to given folder and create it if doesn't exist
        """
        p = Path(folder)
        if not p.exists():
            p.mkdir(parents=True)

        return folder

    def _get_experiment_folder_path(self) -> str:
        """
            Get path to files for this experiment
        """
        return str(
            Path(
                self._experiments_folder, f"{self._branch_name}/{self._experiment_name}"
            )
        )

    def write_summary(self, summary: BaseExperimentSummary):
        """
            Write summary to corresponding file
        """
        file_to_write = self.experiment_content_folder + "/summary.txt"

        with open(file_to_write, "w+") as f:
            print(f"Writing summary to file: {file_to_write}")
            summary_str = str(summary)
            print(summary_str, file=f)  # Print to desired file

    def delete_experiment_content_folder(self):
        """
            Delete experiment content folder
        """
        p = Path(self.experiment_content_folder)

        try:
            shutil.rmtree(p)
        except IOError as e:
            print(f"Couldn't delete folder {str(p)}, error: {e}", file=sys.stderr)

    @property
    def experiment_content_folder(self) -> str:
        """
            Path to folder where experiment content is stored, for example:
                $HOME/.c4v/experiments/<branch_name>/<experiment_name>
        """
        return self._get_experiment_folder_path()

    @property
    def experiments_folder(self) -> str:
        """
            Path to folder where experiments are stored, for example:
                $HOME/.c4v/experiments
        """
        return self._experiments_folder


class BaseExperiment:
    """
        Inherit this class to create new experiments. Essentially, an experiment
        represents some parametrizable process that perform some actions and outputs
        some results. It may use some resources as database managers, scrapers and 
        datasets.

        In order to create an Experiment, you have to inherit this class and implement the following
        functions:
            * experiment_to_run
    """

    def __init__(self, experiment_fs_manager: ExperimentFSManager) -> None:
        self._experiment_fs_manager = experiment_fs_manager

    @property
    def base_folder(self) -> str:
        """
            Folder to store files for this experiment if necessary
        """
        return self._experiment_fs_manager.experiment_content_folder

    def experiment_to_run(self, args: BaseExperimentArguments) -> BaseExperimentSummary:
        """
            Experiment to be ran when run_experiment function is called
            Parameters:
                args : BaseExperimentArguments = Args for the experiment to run
            Return:
                BaseExperimentSummary object summarizing results for this experiment run
        """
        raise NotImplementedError(
            "Should implement experiment_to_ron abstract function"
        )

    def run_experiment(
        self, args: BaseExperimentArguments, delete_after_run: bool = False
    ) -> BaseExperimentSummary:
        """
            Run experiment, givin the given args to the configured experiment, storing its summary
            as a dump in a file in the experiment folder. Delete such folder at the end if requested so.
            Parameters:
                args : BaseExperimentArguments = Args to pass to the experiment when running
                delete_after_run : bool = Should delete experiment folder after run is finished
            Return:
                Experiment result as an Experiment Summary     
        """
        # Run experiment
        summary = self.experiment_to_run(args)

        # Write summary
        self._experiment_fs_manager.write_summary(summary)

        # if nothing else to do, just return
        if not delete_after_run:
            return summary

        # Delete experiment files
        self._experiment_fs_manager.delete_experiment_content_folder()
        return summary

    @classmethod
    def from_branch_and_experiment(cls, branch_name: str, experiment_name: str):
        fs_manager = ExperimentFSManager(branch_name, experiment_name)
        return cls(fs_manager)



"""
    This classifier class will perform basic experiments, receiving 
    as arguments the training arguments and the columns to use from the training dataset
"""
# Local imports
import dataclasses
from c4v.config import settings
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData, Labels
from c4v.classifier.base_model import BaseModel, C4vDataFrameLoader

# Python imports
from typing import Dict, List, Any, Tuple
from pathlib import Path
from pandas.core.frame import DataFrame
from importlib import resources
from enum import Enum

# Third Party
from pandas.core.frame import DataFrame
from transformers import (
    RobertaTokenizer,
    Trainer,
    TrainingArguments,
    RobertaForSequenceClassification,
    AutoModelForSequenceClassification,
    AutoTokenizer,
)
from transformers_interpret import SequenceClassificationExplainer
from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score
from sklearn.model_selection import train_test_split
from datasets import Dataset
import torch
import pandas as pd
import numpy as np

from transformers.trainer_utils import EvalPrediction

BASE_C4V_FOLDER = settings.c4v_folder
BASE_LANGUAGE_MODEL = settings.default_base_language_model


class Classifier(BaseModel):
    """
        This class provides a simple way to run simple experiments.
    """

    def get_dataframe(self, dataset_name: str) -> DataFrame:
        """
            Get dataframe as a pandas dataframe, using a csv file stored in <project_root>/data/processed/huggingface
        """
        return C4vDataFrameLoader.get_from_processed(dataset_name)

    def prepare_dataframe(
        self, columns: List[str], dataset_name: str
    ) -> Tuple[List[str], List[int]]:
        """
            Return the list of text bodies and its corresponding label of whether it is 
            a missing service problem or not, expressed as int
            Parameters:
                columns : [str] = List of columns to use as part of the experiment 
                dataset_name : str = name of the dataset to use
            Return:
                ([str], [int]) = the i'st position of the first list is the body of a news article, and the 
                                 i'st position of the second list tells whether the article i talks about
                                 a missing service or not, expressed as an int (1 if it is, 0 if not)
        """

        df_pscdd = self.get_dataframe(dataset_name).sample(
            frac=1
        )  # use sample to shuffle rows

        df_pscdd["label"] = (
            df_pscdd["label"].apply(lambda x: "IRRELEVANTE" not in x)
        ).astype(int)

        df_pscdd = df_pscdd.convert_dtypes()
        df_issue_text = df_pscdd[[*columns, "label"]]
        df_issue_text.dropna(inplace=True)

        x = [
            "\n".join(tup)
            for tup in zip(*[list(df_issue_text[col]) for col in columns])
        ]

        y = list(df_issue_text["label"])

        return x, y

    def load_tokenizer(self, model_name: str = None) -> RobertaTokenizer:
        """
            Create & configure tokenizer from hub
            Parameters:
                model_name : str = (optional) name of the model whose tokenizer will be loaded, defaults to internally configured one
            Return:
                RobertaTokenizer: tokenizer to retrieve
        """
        return AutoTokenizer.from_pretrained(
            model_name or self._base_model_name, id2label=self.get_id2label_dict()
        )

    def load_base_model(
        self, model_name: str = None
    ) -> RobertaForSequenceClassification:
        """
            Create model from model hub, configure them and retrieve it
            Parameters:
                model_name : str = (optional) name of the model to load
            Return:
                RobertaForSequenceClassification : the model as specified
        """
        # Creating model and tokenizer
        model = AutoModelForSequenceClassification.from_pretrained(
            model_name or self._base_model_name, num_labels=2
        )
        # Use GPU if available
        model.to(self._device)

        return model

    def transform_dataset(
        self, x: List[str], y: List[int], tokenizer: RobertaTokenizer,
    ) -> Dataset:
        """
            perform operations needed to post process a dataset separated in input list
            "x" and expected answers list "y", returning a training and a validation datasets,
            ready to be fed into the model 
            Parameters:
                x : [str] = List of content inputs to train with
                y : [int] = List of integer values matching answer for inputs in x list
            Return:
                (Dataset, Dataset) = the training and the validation dataset, in such order
        """

        X_tokenized = tokenizer(x, padding=True, truncation=True, max_length=512)

        # Create torch dataset
        class _Dataset(torch.utils.data.Dataset):
            def __init__(self, encodings, labels=None):
                self.encodings = encodings
                self.labels = labels

            def __getitem__(self, idx):
                item = {
                    key: torch.tensor(val[idx]) for key, val in self.encodings.items()
                }
                if self.labels:
                    item["labels"] = torch.tensor(self.labels[idx])
                return item

            def __len__(self):
                return len(self.encodings["input_ids"])

        dataset = _Dataset(X_tokenized, y)

        return dataset

    @staticmethod
    def compute_metrics(predictions: EvalPrediction) -> Dict[str, Any]:
        """
            Use this function to evaluate metrics for a given prediction 
            Parameters:
                predictions : EvalPrediction = Prediction provided by the model with certain input 
            Return:
                Dict with computed metrics. Provided fields:
                    - accuracy
                    - precision
                    - recall
                    - f1
        """
        pred, labels = predictions
        pred = np.argmax(pred, axis=1)

        accuracy = accuracy_score(y_true=labels, y_pred=pred)
        recall = recall_score(y_true=labels, y_pred=pred)
        precision = precision_score(y_true=labels, y_pred=pred)
        f1 = f1_score(y_true=labels, y_pred=pred)

        return {
            "accuracy": accuracy,
            "precision": precision,
            "recall": recall,
            "f1": f1,
        }

    def _get_default_train_args(self) -> Dict[str, Any]:
        """
            Return a default version of training arguments as a dict
        """

        default = {
            "output_dir": self.results_path,
            "num_train_epochs": 1,
            "per_device_train_batch_size": 10,
            "per_device_eval_batch_size": 10,
            "warmup_steps": 500,
            "weight_decay": 0.01,
            "logging_dir": self.logs_path,
            "save_total_limit": 1,
        }

        return default

    def _override_train_args(self, new_args: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
            Return the default args dict, with overriden settings specified as the ones specified in 
            input dict
        """

        default = self._get_default_train_args()
        for (k, v) in new_args.items():
            default[k] = v

        return default

    def train_and_save_model(
        self,
        model: Any,
        output_dir: str = None,
        logging_dir: str = None,
        path_to_save_checkpoint: str = None,
        train_args: Dict[str, Any] = None,
        train_dataset: Dataset = None,
        eval_dataset: Dataset = None,
    ) -> Trainer:
        """
            Train and save using provided model, using provided args and storing 
            results to "output_dir" and logging additional info to "logging_dir",
            saving checkpoint to "path_to_save_checkpoint". If some of this values is not provided,
            it will be defaulted to this experiment's dedicated folder 
            Parameters:
                model = model to train
                output_dir : str = results folder, defaulted to experiment's results folder
                logging_dir : str = logs folder, defaulted to experiment's logs folder
                path_to_save_checkpoint : str = path to save checkpoint after training is finished, defaults to experiment's folder
                train_args : TrainArguments = if provided, override all default parameters passed to the training
            Return:
                properly configured trainer instance

        """
        if output_dir:
            train_args["output_dir"] = output_dir
        if logging_dir:
            train_args["logging_dir"] = logging_dir

        args = TrainingArguments(**self._override_train_args(train_args))

        trainer = Trainer(
            model=model,
            args=args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            compute_metrics=self.__class__.compute_metrics,
        )

        # Train pre-trained model
        trainer.train()

        model.save_pretrained(path_to_save_checkpoint or self.files_folder_path)

        return trainer

    def load_fine_tuned_model(
        self, path: str = None
    ) -> RobertaForSequenceClassification:
        """
            load fine tuned model from provided path, defaults to an already trained one 
            in experiment's folder
        """
        path = path or self._files_folder

        # Check if path is a valid one
        if not Path(path, "config.json").exists():
            raise ValueError(f"Experiment does not exists: {path}")

        model = AutoModelForSequenceClassification.from_pretrained(
            path, local_files_only=True, id2label=self.get_id2label_dict()
        )
        return model

    def evaluate_metrics(self, trainer: Trainer, val_dataset: Dataset) -> DataFrame:
        """
            Compute metrics as a dataframe for this trainer object using the validation dataset
            Parameters:
                trainer : Trainer = trainer object already trained and ready to test
                val_dataset : Dataset = data used for validation
            Return:
                A dataframe with output of a validation
        """
        ## Evaluate Metrics
        metrics = trainer.evaluate(val_dataset)
        metrics_df = pd.DataFrame.from_dict(
            metrics, orient="index", columns=["metrics_value"]
        )
        return metrics_df

    def run_training(
        self,
        train_args: Dict[str, Any] = None,
        columns: List[str] = ["content"],
        training_dataset: str = "classifier_training_dataset.csv",
        confirmation_dataset: str = "classifier_confirmation_dataset.csv",
        val_test_proportion: float = 0.2,
        base_model_name: str = None,
    ) -> DataFrame:
        """
            Run an experiment specified by given train_args, and write a summary if requested so
            Parameters:
                train_args : Dict[str, Any] = (optional) arguments passed to trainig arguments
                columns : [str] = (optional) columns to use in the dataset
                training_dataset : str = (optional) dataset name to use during training, should be a name of a dataset under <project_root>/data/raw/huggingface
                confirmation_dataset : str = (optional) dataset name to use after training as confirmation dataset
                val_test_proportion : float = (optional) how much proportion of the training dataset to use as validation dataset
                base_model_name : str = (optional) Name of the model to use as a base for this experiment, defaults to the stored one if not provided
            Return:
                Classifier metrics
        """

        # Sanity checks
        assert (
            0.0 < val_test_proportion < 1.0
        ), "val_test_proportion should be in range (0,1)"
        assert all(
            c in [f.name for f in dataclasses.fields(ScrapedData)] for c in columns
        ), "columns should be valid ScrapedData fields"

        # check that you have a folder where to store results
        if not self.files_folder_path:
            raise ValueError(
                "Can't train in a Classifier without a folder for local data"
            )

        # Prepare training dataframe and load model + tokenizer
        x, y = self.prepare_dataframe(columns=columns, dataset_name=training_dataset)

        # Split dataset into training and validation
        X_train, X_val, y_train, y_val = train_test_split(
            x, y, test_size=val_test_proportion
        )

        # Load model and tokenizer
        model = self.load_base_model(model_name=base_model_name)
        tokenizer = self.load_tokenizer(model_name=base_model_name)

        # transform data into datasets
        train_dataset = self.transform_dataset(X_train, y_train, tokenizer)
        val_dataset = self.transform_dataset(X_val, y_val, tokenizer)
        del X_train, X_val, y_train, y_val

        # Fine tune the model
        fine_tuned_model_trainer = self.train_and_save_model(
            model=model,
            output_dir=self.results_path,
            logging_dir=self.logs_path,
            path_to_save_checkpoint=self.files_folder_path,
            train_args=self._override_train_args(train_args or {}),
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
        )

        del train_dataset, val_dataset

        # Save tokenizer too
        tokenizer.save_pretrained(self.files_folder_path)

        # Prepare confirmation dataframe
        x_confirmation, y_confirmation = self.prepare_dataframe(
            columns=columns, dataset_name=confirmation_dataset
        )
        confirmation_dataset = self.transform_dataset(
            x_confirmation, y_confirmation, tokenizer
        )

        # Get the metrics from the model
        metrics_df = self.evaluate_metrics(
            trainer=fine_tuned_model_trainer, val_dataset=confirmation_dataset
        )

        return metrics_df

    def _get_text_from_scrapeddata(self, data: ScrapedData) -> str:
        """
            Returns a string constructed from a scraped data instance
        """
        return data.title

    def classify(
        self, data: List[ScrapedData], model: str = None
    ) -> List[Dict[str, Any]]:
        """
            Classify the given data instance, returning classification metrics
            as a simple dict.
            Parameters:
                data : ScrapedData = Data instance to classify
                model : str = model name of model to load use when classifying. If no model provided,
                              use the model configured for this classifier
            Return:
                A List of dicts with the resulting scraped data correctly labelled
                and its corresponding scores tensor for each possible label. Available fields:
                    + data : ScrapedData = resulting data instance after classification
                    + scores : torch.Tensor = Scores for each label returned by the classifier
        """
        # Get model from experiment:
        if model is None:
            model = self._files_folder
            # Check if there's a config.json in this folder
            if not Path(model, "config.json").exists():
                raise ValueError(f"path: {model} does not contains valid model to load")

        roberta_model = self.load_fine_tuned_model(model)

        # Tokenize input
        roberta_tokenizer = self.load_tokenizer()
        tokenized_input = roberta_tokenizer(
            [self._get_text_from_scrapeddata(d) for d in data],
            padding=True,
            truncation=True,
            max_length=512,
            return_tensors="pt",
        )

        raw_output = roberta_model(**tokenized_input)
        output: torch.Tensor = torch.nn.functional.softmax(raw_output.logits, dim=-1)

        result = []
        for (x, d) in zip(output, data):
            d.label = Labels(self.index_to_label(torch.argmax(x).item()))
            result.append({"data": d, "scores": x})

        return result

    def explain(
        self, sentence: str, html_file: str = None, additional_label: str = None
    ) -> Dict[str, Any]:
        """
            Return a list of words from provided sentence with how much they collaborate to each label 
            Parameters:
                sentence : str = text to explain
                html_file : str = path to some html file to store human readable representation. If no provided, 
                                    it's ignored
                additional_label : str = Label to include in expalantion. If the predicted label is different 
                                         from this one, then explain how much this label was contributing to 
                                         its corresponding value. Ignored if not provided.
            Return:
                Dict with data for this explanation. For example:
                {   "scores" : 
                    [   
                        ('denuncian' , 0.98932),
                        ('falta'     , 0.78912),
                        ('de'        , 0.001231),
                        ('agua'      , 0.863781)
                    ],
                    "label" : "DENUNCIA_FALTA_DEL_SERVICIO",
        """
        # Load model and tokenizer
        model = self.load_fine_tuned_model()
        tokenizer = self.load_tokenizer()

        # Create explainer
        explainer = SequenceClassificationExplainer(model, tokenizer)

        scores = explainer(sentence, class_name=additional_label)
        label = explainer.predicted_class_name

        # write html file
        if html_file:
            explainer.visualize(html_file)

        return {"scores": scores, "label": label}

    def get_id2label_dict(self) -> Dict[int, str]:
        """
            Return dict mapping from ids to labels
        """
        return {
            1: Labels.DENUNCIA_FALTA_DEL_SERVICIO.value,
            0: Labels.IRRELEVANTE.value,
        }

    def index_to_label(self, index: int) -> Labels:
        """
            Get index for label
        """
        d = self.get_id2label_dict()

        return d.get(index, Labels.IRRELEVANTE.value)

    @staticmethod
    def get_labels() -> List[str]:
        """
            Get list of possible labels outputs
        """
        return Labels.labels()



"""
    Base class for scrapy-based scrapers.

    In order to create a a new scrapy scraper:
        1) Create a new scraper un "scrapers" folder, and make it subclass
            of this BaseScrapyScraper
        2) override "spider" attribute of your new class with a valid
            scrapy spider
        3) wired it in settings as you would do with a regular scraper 
"""

# External imports
from scrapy import Spider

# Internal imports
from c4v.scraper.scrapers.base_scraper import BaseScraper
from c4v.scraper.spider_manager import SpiderManager
from c4v.scraper.scraped_data_classes.base_scraped_data import BaseDataFormat

# Python imports
from typing import Type, List


class BaseScrapyScraper(BaseScraper):
    """
        In order to create a new Scrappy Scrapper, just 
        inherit this class and assign a new value to the 
        "spider" field, a valid scrapy Spider sub class.
    """

    spider: Type[Spider] = None

    def __init__(self):

        if self.spider is None:
            raise TypeError(
                "Spider not defined,"
                + " perhaps you forgot to override spider"
                + " attribute in BaseScrapyScraper subclass?"
            )

        self._spider_manager = SpiderManager(self.spider)

    def parse(self, response) -> BaseDataFormat:
        return self._spider_manager.parse(response)

    def scrape(self, url: str) -> BaseDataFormat:
        return self._spider_manager.scrape(url)

    def schedule_scraping(self, urls: List[str]):
        return self._spider_manager.schedule_scraping(urls)

    def start_bulk_scrape(self):
        return self._spider_manager.start_bulk_scrape()

    def get_scraped_items(self) -> List[BaseDataFormat]:
        return self._spider_manager.get_scraped_items()



"""
    Class for scraping primicia data
"""
from c4v.scraper.crawler.crawlers.base_crawler import BaseCrawler


class PrimiciaCrawler(BaseCrawler):
    """
        Class to crawl Primicia Urls
    """

    start_sitemap_url = "https://primicia.com.ve/sitemap_index.xml"
    name = "primicia"

    # URL patterns for non relevant urls (news that have nothing to do with public services)
    IRRELEVANT_URLS = [
        ".*primicia.com.ve/mundo/.*",
        ".*primicia.com.ve/deportes/.*",
        ".*primicia.com.ve/placeres/.*",
        ".*primicia.com.ve/mas/virales/.*",
        ".*primicia.com.ve/mas/ciencia-y-tecnologia/.*",
        ".*primicia.com.ve/sucesos/.*",
        ".*primicia.com.ve/politica/.*",
    ]

    @staticmethod
    def should_crawl(url: str) -> bool:
        # Sitemaps about posts will start with this prefix
        return url.startswith("https://primicia.com.ve/post-sitemap")

    @staticmethod
    def should_scrape(url: str) -> bool:
        # Checks if provided url starts with base site
        # and if its length is creater to base site (so we avoid crawling main page
        # by accident)
        primicia = "https://primicia.com.ve/"
        return url.startswith(primicia) and len(url) > len(primicia)



"""
    The manager may need to store some persistent configuration data
    between runs, here we find the scheme requested for the system
"""
# Local imports
from c4v.config import settings

# Python imports
import dataclasses
import json


@dataclasses.dataclass
class Metadata:
    """
        Data required and stored by the manager to perform common operations
    """

    classifier_model: str = settings.default_base_language_model  # Abosolute path for the model to load
    base_language_model: str = settings.default_base_language_model  # Abosolute path for the base model to load

    def to_json_str(self, pretty: bool = False) -> str:
        """
            Return a json-formated string representation for this object
        """
        indent = 1 if pretty else None
        return json.dumps(self, cls=_MetadataJSONEncoder, indent=indent)

    @classmethod
    def from_json_str(cls, json_str: str):
        """
            Create instance from json-formated string
        """
        return cls(**json.loads(json_str))

    def to_json(self, filename: str, pretty: bool = False) -> str:
        """
            Return a json-formated string representation for this object
        """
        indent = 1 if pretty else None
        with open(filename, "w") as f:
            return json.dump(self, f, cls=_MetadataJSONEncoder, indent=indent)

    @classmethod
    def from_json(cls, json_path: str):
        """
            Create an instance from a json file
            Parameters:
                json_path : str = path to file to load, may raise FileNotFound error if file does not exists
        """
        with open(json_path) as f:
            return cls(**json.load(f))


class _MetadataJSONEncoder(json.JSONEncoder):
    """
        Class to serialize metadata into a json file
    """

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)



"""
    Base class for a scrapper.
    In order to create and wire a new scrapper:
        1) Create a new scraper in the "scrapers" directory
        2) Make your scraper a subclass of BaseScraper
        3) Implement missing methods (parse & scrape)
        4) add an entry in settings.py to the INSTALLED_SCRAPERS list. 
            Import it if necessary
"""

# Local imports
from c4v.scraper.scraped_data_classes.base_scraped_data import BaseDataFormat

# Python imports
from typing import List


class BaseScraper:
    """
        Base class for scrapers implementations
    """

    # domain to be scraped by this scraper
    intended_domain: str = None

    def parse(self, response) -> BaseDataFormat:
        """
            return scraped data from a response object 
            Parameters:
                + response : any = some kind of structure holding an http response
                                   from which we can scrape data
            Return:
                A dict with scrapped fields from response
        """
        raise NotImplementedError("Scrapers should implement the parse method")

    def scrape(self, url: str) -> BaseDataFormat:
        """
            return scraped data from url.
            Parameters: 
                + url : str = url to be scraped by this class
            Return:
                A dict with scrapped data from the given url
                if such url is a valid one
        """
        raise NotImplementedError("Scrapers should implement the scrape method")

    def bulk_scrape(self, urls: List[str]) -> List[BaseDataFormat]:
        """
            Return scraped data for a list of urls. Override it 
            if your scraper implementation could handle an optimized
            bulk scraping.

            Parametes:
                + urls : [str] = urls to be scraped
            Return:
                List of scraped items. Notice that the order it's not guaranteed to be
                the same as in the input list.
        """
        self.schedule_scraping(urls)
        self.start_bulk_scrape()
        return self.get_scraped_items()

    def schedule_scraping(self, urls: List[str]):
        """
            Schedule a list of urls to be scraped later.
            This is necessary so you can scrape data from multiple
            sources at the same time with multiple scrapers.

            Parameters:
                + urls : [str] = urls to be scheduled to scrape
        """
        self._to_scrape = urls

    def start_bulk_scrape(self):
        """
            Start a bulk scraping process, storing results internally
        """

        # Consistency check: cannot bulk scrape without something to scrape
        assert hasattr(
            self, "_to_scrape"
        ), "There's no scheduled items to scrape. Call schedule_scraping first to set items to be scraped"
        items = []

        for url in self._to_scrape:
            item = self.scrape(url)
            if item:
                items.append(item)

        del self._to_scrape
        self._scraped_items = items

    def get_scraped_items(self) -> List[BaseDataFormat]:
        """
            Return the scraped list of items after a bulk scrape process
        """
        assert hasattr(
            self, "_scraped_items"
        ), "There's no scraped items. Call start_bulk_scrape first to get items to retrieve"

        out = self._scraped_items
        del self._to_scrape

        return out



"""
    Creates a confirmation dataset, using a list of well-known urls for news that have problems.
    Expected fields for such csv:
        + url
        + label
    Arguments:
        valid_dataset_filename : str = path to file to load as a dataframe containing only urls to news with positive labels
        train_dataset_filename : str = path to training dataset, used to know which urls won't be used for confirmation dataset
        db_name : str = (optional) filename of the db used to retrieve non-relevant urls, by default using the local sqlite db
        n_irrelevant_rows : int = (optional) amount of rows to be used for irrelevant news
"""
# Local imports
from typing import List
from c4v.config import settings
from c4v.microscope.manager import Manager
from c4v.scraper.crawler.crawlers.primicia_crawler import PrimiciaCrawler

# Third party imports
import pandas as pd
import sys

# Ammount of arguments for this script
_N_ARGS = 2

# Sanity check: Enough arguments
if len(sys.argv) < (_N_ARGS + 1):
    print("[ERROR] Not enough arguments", file=sys.stderr)

# Read arguments from command line
# Read positive labels dataset
valid_dataset_filename: str = sys.argv[1]
print(f"Reading positive labels dataset: {valid_dataset_filename}...")
valid_dataset_df = pd.read_csv(valid_dataset_filename)
valid_dataset_df.dropna(axis=0, inplace=True, subset=["url"])

# Read training dataset
train_dataset_filename: str = sys.argv[2]
print(f"Reading training  dataset: {train_dataset_filename}...")
train_dataset_df = pd.read_csv(train_dataset_filename)

# Read sqlite db argument
db_name = sys.argv[3] if len(sys.argv) > (_N_ARGS + 1) else settings.local_sqlite_db

# Read how much irrelevant rows to use
n_irrelevant_rows: int = int(sys.argv[3]) if len(sys.argv) > (_N_ARGS + 1) else 10000
print(
    f"Will try to retrieve {n_irrelevant_rows} irrelevant rows from database {db_name}..."
)

# Set up manager
m: Manager = Manager.from_local_sqlite_db(db_name)

# -- Collecting urls to scrape -----------------------------------

# Get urls to scrape as problems
positive_urls_to_scrape, non_scrapable = m.split_non_scrapable(
    list(valid_dataset_df["url"])
)

if non_scrapable:  # warn about non-scrapable urls
    print(
        f"[WARNING] {len(non_scrapable)} urls from {len(valid_dataset_df)} won't be scraped, as they're not scrapable",
        file=sys.stderr,
    )
    del non_scrapable

# Get data not in training dataset
not_valid_urls = set(train_dataset_df["url"]).union(positive_urls_to_scrape)
del train_dataset_df

print(f"Crawling irrelevant data...")
# Crawl data for primicia
primicia_crawler = PrimiciaCrawler.from_irrelevant()
# crawled urls
irrelevant_urls: List[str] = []


def collect(crawled_urls: List[str]):
    irrelevant_urls.extend(url for url in crawled_urls if url not in not_valid_urls)
    print(f"Collected items: {len(irrelevant_urls)} / {n_irrelevant_rows}")


def should_stop() -> bool:
    return len(irrelevant_urls) >= n_irrelevant_rows


irrelevant_urls = irrelevant_urls[:n_irrelevant_rows]

# Crawl items
primicia_crawler.crawl_and_process_urls(
    post_process_data=collect, should_stop=should_stop
)

print(f"Collected irrelevant articles: {len(irrelevant_urls)}")

# Scrape crawled items

# -- Scraping urls -------------------------------
# Create positive dataframe
print(f"Scraping collected items...")
scraped_data = m.get_bulk_data_for(
    positive_urls_to_scrape + irrelevant_urls
)  # get data
scraped_positive_data_df = pd.DataFrame(
    x for x in scraped_data if x.url in positive_urls_to_scrape
)  # turn into dataset
# remove unnecesary rows:
valid_dataset_df.drop(
    [c for c in valid_dataset_df.columns if c != "label" and c != "url"],
    inplace=True,
    axis=1,
)
scraped_positive_data_df = pd.merge(  # join dataframes to get corresponding labels
    scraped_positive_data_df, valid_dataset_df, "inner", on="url"
)

# Create Irrelevant dataframe
irrelevant_scraped_df = pd.DataFrame(
    x for x in scraped_data if x.url in irrelevant_urls
)

# Remove irrelevant cols
irrelevant_scraped_df.drop(
    [
        c
        for c in irrelevant_scraped_df.columns
        if c not in scraped_positive_data_df.columns
    ],
    inplace=True,
    axis=1,
)
irrelevant_scraped_df["label"] = [["IRRELEVANTE"]] * len(irrelevant_scraped_df)

full_df = pd.concat(
    [scraped_positive_data_df, irrelevant_scraped_df], ignore_index=True
)
full_df = full_df.sample(frac=1.0)
print(full_df)
print("Creating final dataset...")

filename = "confirmation_dataset.csv"
print(f"Saving file to: {filename}")
full_df.to_csv(filename)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['c4v',
 'c4v.classifier',
 'c4v.classifier.data_gathering',
 'c4v.classifier.language_model',
 'c4v.data',
 'c4v.microscope',
 'c4v.scraper',
 'c4v.scraper.crawler',
 'c4v.scraper.crawler.crawlers',
 'c4v.scraper.persistency_manager',
 'c4v.scraper.scraped_data_classes',
 'c4v.scraper.scrapers',
 'c4v.scraper.spiders']

package_data = \
{'': ['*']}

install_requires = \
['Scrapy>=2.5.0,<3.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'click==7.1.2',
 'datasets>=1.10.2,<2.0.0',
 'dynaconf>=3.1.4,<4.0.0',
 'google-cloud-bigquery==1.26.1',
 'importlib-metadata>=4.6.1,<5.0.0',
 'importlib-resources>=5.2.2,<6.0.0',
 'ipykernel==5.5.5',
 'ipython==7.16.1',
 'jupyter>=1.0.0,<2.0.0',
 'nbconvert==5.6.1',
 'nltk>=3.5,<4.0',
 'pandas==1.1.1',
 'pip>=21.0.0,<22.0.0',
 'scikit-learn==0.23.1',
 'scikit-multilearn==0.2.0',
 'scipy==1.5.4',
 'tabulate>=0.8.9,<0.9.0',
 'tensorflow-probability==0.10.0',
 'tensorflow==2.5.0',
 'tensorflow_hub[make_image_classifier]==0.8.0',
 'torch>=1.9.0,<1.10.0',
 'traitlets==4.3.3',
 'transformers-interpret>=0.5.2,<0.6.0',
 'transformers>=4.9.0,<5.0.0',
 'zipp>=3.5.0,<4.0.0']

extras_require = \
{':python_full_version >= "3.6.1" and python_version < "3.7"': ['dataclasses>=0.8.0,<0.9.0']}

entry_points = \
{'console_scripts': ['c4v = c4v.c4v_cli:c4v_cli']}

setup_kwargs = {
    'name': 'c4v-py',
    'version': '0.1.0.dev202111011437',
    'description': 'Code for Venezuela python library.',
    'long_description': '# c4v-py\n\n<p align="center">\n  <img width="125" src="assets/logo.png">\n</p>\n\n> Solving Venezuela pressing matters one commmit at a time\n\n`c4v-py` is a library used to address Venezuela\'s pressing issues\nusing computer and data science.\n\n- [Installation](#installation)\n- [Development](#development)\n- [Pending](#pending)\n\n## Installation\n\nUse pip to install the package:\n\n```python3\npip install c4v-py\n```\n\n## Usage\n\n_TODO_\n\n[Can you help us? Open a new issue in\nminutes!](https://github.com/code-for-venezuela/c4v-py/issues/new/choose)\n\n## Contributing\n\nThe following tools are used in this project:\n\n- [Poetry](https://python-poetry.org/) is used as package manager.\n- [Nox](https://nox.thea.codes/) is used as automation tool, mainly for testing.\n- [Black](https://black.readthedocs.io/) is the mandatory formatter tool.\n- [PyEnv](https://github.com/pyenv/pyenv/wiki) is recommended as a tool to handle multiple python versions in your machine.\n\nThe library is intended to be compatible with python ~3.6.9, ~3.7.4 and ~3.8.2. But the primary version to support is ~3.8.2.\n\nThe general structure of the project is trying to follow the recommendations\nin [Cookiecutter Data Science](https://drivendata.github.io/cookiecutter-data-science/).\nThe main difference lies in the source code itself which is not constraint to data science code.\n\n### Setup\n\n1. Install pyenv and select a version, ie: 3.8.2. Once installed run `pyenv install 3.8.2`\n2. Install poetry in your system\n3. Clone this repo in a desired location `git clone https://github.com/code-for-venezuela/c4v-py.git`\n4. Navigate to the folder `cd c4v-py`\n5. Make sure your poetry picks up the right version of python by running `pyenv local 3.8.2`, if 3.8.2 is your right version.\n6. Since our toml file is already created, we need to get all dependencies by running `poetry install`. This step might take a few minutes to complete.\n7. Install nox\n8. From `c4v-py` directory, on your terminal, run the command `nox -s tests` to make sure all the tests run.\n\nIf you were able to follow every step with no error, you are ready to start contributing. Otherwise, [open a new issue](https://github.com/code-for-venezuela/c4v-py/issues/new/choose)!\n\n## Roadmap\n\n- [ ] Add CONTRIBUTING guidelines\n- [ ] Add issue templates\n- [ ] Document where to find things (datasets, more info, etc.)\n  - This might be done (in conjunction) with Github Projects. Managing tasks there might be a good idea.\n- [ ] Add LICENSE\n- [ ] Change the authors field in pyproject.toml\n- [ ] Change the repository field in pyproject.toml\n- [ ] Move the content below to a place near to the data in the data folder or use the reference folder.\n      Check [Cookiecutter Data Science](https://drivendata.github.io/cookiecutter-data-science/) for details.\n- [ ] Understand what is in the following folders and decide what to do with them.\n  - [ ] brat-v1.3_Crunchy_Frog\n  - [ ] creating_models\n  - [x] data/data_to_annotate\n  - [ ] data_analysis\n- [ ] Set symbolic links between `brat-v1.3_Crunchy_Frog/data` and `data/data_to_annotate`. `data_sampler` extracts to `data/data_to_annotate`. Files placed here are read by Brat.\n  - [ ] Download Brat - `wget https://brat.nlplab.org/index.html`\n  - [ ] untar brat - `tar -xzvf brat-v1.3_Crunchy_Frog.tar.gz`\n  - [ ] install brat - `cd brat-v1.3_Crunchy_Frog && ./install.sh`\n  - [ ] replace default annotation conf for current configuration - `wget https://raw.githubusercontent.com/dieko95/c4v-py/master/brat-v1.3_Crunchy_Frog/annotation.conf -O annotation.conf`\n  - [ ] replace default config.py for current configuration - `wget https://raw.githubusercontent.com/dieko95/c4v-py/master/brat-v1.3_Crunchy_Frog/config.py -O config.py`\n',
    'author': 'Edilmo Palencia',
    'author_email': 'edilmo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.codeforvenezuela.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)



"""
    This file will take a list of csv files and merge them in a single one 
    ready for training
"""
import pandas as pd
import sys

files = sys.argv[1:]

# Check if no files to work with
if not files:
    print("No files to merge")
    exit(0)

# Open every file and load dataframes
dfs = []
for file in files:
    print(f"Openning file: {file}...")
    with open(file) as f:
        new_df: pd.DataFrame = pd.read_csv(f)
    print(f"Rows in file: {len(new_df)}")
    dfs.append(new_df)

# Remove irrelevant columns
columns = ["url", "title", "content", "author", "categories", "date", "label"]

# Concate dataframes
df = pd.concat(dfs)
del dfs

# Remove irrelevant columns
columns_to_remove = [col for col in df.columns if col not in columns]
print(f"Removing columns: {columns_to_remove}")

print("Resulting dataframe: ")
print(df)


# Saving dataframe
filename = "aggregated_dataframe.csv"
print(f"Saving csv to: {filename}")
df.to_csv(filename)
